{------------------------------------------------------------------------------

Copyright (C) 2002  Rainer Queck
e-mail: Rainer.Queck@Qutronic.de
http://www.qutronic.de


This code is free; you can redistribute it and/or
modify it under the terms of the MPL MOZILLA PUBLIC LICENSE. 

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
MPL MOZILLA PUBLIC LICENSE for more details at http://www.mozilla.org/MPL

//----------------------------------------------------------------------------}
// Unit Created at    : somewhen in 2002
// Unit last modified : Date: 21.12.2002  Time :10:13:30
//------------------------------------------------------------------------------
unit AutoDiagramming;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, MMToolsApi,MMDiagramApi,MMEngineDefs, RQClasses;

type
  TMsgsInfo=record
    NumberOfMessages,
    NumberOfSelfCalls : Integer;
  end;
  TSqDgRole = class;
  TSqDgMsg = class;
  TSqRoleList = class;
  TSqMsgList = class;
  TSequenceDiagram = class;
  TDiagramGenerator = class;
  TSqDgRole = class (TObject)
  private
    FClassBase: IMMClassBase;
    FParent: TSequenceDiagram;
    FPositionX: Integer;
    FPositionY: Integer;
    FSymbol: IMMSymbol;
    function GetAutoSizeX: Boolean;
    function GetAutoSizeY: Boolean;
    function GetHeight: Integer;
    function GetInstanceName: string;
    function GetName: string;
    function GetRoleKind: string;
    function GetWidth: Integer;
    procedure SetAutoSizeX(Value: Boolean);
    procedure SetAutoSizeY(Value: Boolean);
    procedure SetHeight(Value: Integer);
    procedure SetInstanceName(const Value: string);
    procedure SetName(const Value: string);
    procedure SetRoleKind(const Value: string);
    procedure SetWidth(Value: Integer);
  public
    constructor Create(Parent : TSequenceDiagram);
    destructor Destroy; override;
    procedure AddExternalDocHyperlink(aHyperlink : String);
    procedure HotLinkToEntity(aEntity : TLinkableEntity; aId : Integer);
    procedure Move(xPos,yPos : integer);
    property AutoSizeX: Boolean read GetAutoSizeX write SetAutoSizeX;
    property AutoSizeY: Boolean read GetAutoSizeY write SetAutoSizeY;
    property ClassBase: IMMClassBase read FClassBase write FClassBase;
    property Height: Integer read GetHeight write SetHeight;
    property InstanceName: string read GetInstanceName write SetInstanceName;
    property Name: string read GetName write SetName;
    property PositionX: Integer read FPositionX write FPositionX;
    property PositionY: Integer read FPositionY write FPositionY;
    property RoleKind: string read GetRoleKind write SetRoleKind;
    property Symbol: IMMSymbol read FSymbol write FSymbol;
    property Width: Integer read GetWidth write SetWidth;
  end;
  
  TSqDgMsg = class (TObject)
  private
    FAssociation: IMMAssociation;
    FCondition: string;
    FEndTime: string;
    FMessageKind: string;
    FName: string;
    FParameters: string;
    FParent: TSequenceDiagram;
    FSourceRole: TSqDgRole;
    FStartTime: string;
    FTargetRole: TSqDgRole;
    function GetEndTime: Integer;
    function GetMessageKind: string;
    function GetName: string;
    function GetParameters: string;
    function GetStartTime: Integer;
    procedure SetCondition(const Value: string);
    procedure SetEndTime(Value: Integer);
    procedure SetMessageKind(const Value: string);
    procedure SetName(const Value: string);
    procedure SetParameters(const Value: string);
    procedure SetStartTime(const Value: Integer);
  public
    constructor Create(Parent : TSequenceDiagram; Source,Target : TSqDgRole);
    destructor Destroy; override;
    procedure Commit;
    property Association: IMMAssociation read FAssociation write FAssociation;
    property Condition: string read FCondition write SetCondition;
    property EndTime: Integer read GetEndTime write SetEndTime;
    property MessageKind: string read GetMessageKind write SetMessageKind;
    property Name: string read GetName write SetName;
    property Parameters: string read GetParameters write SetParameters;
    property SourceRole: TSqDgRole read FSourceRole write FSourceRole;
    property StartTime: Integer read GetStartTime write SetStartTime;
    property TargetRole: TSqDgRole read FTargetRole write FTargetRole;
  end;
  
  TSqRoleList = class (TList)
  private
    function GetRole(Index: Integer): TSqDgRole;
    procedure SetRole(Index: Integer; Value: TSqDgRole);
  public
    constructor Create;
    function Add(Item: TSqDgRole): Integer;
    function Extract(Item: TSqDgRole): TSqDgRole;
    function First: TSqDgRole;
    function IndexOf(Item: TSqDgRole): Integer;
    procedure Insert(Index: Integer; Item: TSqDgRole);
    function Last: TSqDgRole;
    function Remove(Item: TSqDgRole): Integer;
    procedure Sort;
    property Items[Index: Integer]: TSqDgRole read GetRole write SetRole;
  end;
  
  TSqMsgList = class (TList)
  private
    function GetMessage(Index: Integer): TSqDgMsg;
    procedure SetMessage(Index: Integer; Value: TSqDgMsg);
  public
    constructor Create;
    function Add(Item: TSqDgMsg): Integer;
    function Extract(Item: TSqDgMsg): TSqDgMsg;
    function First: TSqDgMsg;
    function IndexOf(Item: TSqDgMsg): Integer;
    procedure Insert(Index: Integer; Item: TSqDgMsg);
    function Last: TSqDgMsg;
    function Remove(Item: TSqDgMsg): Integer;
    procedure Sort;
    property Items[Index: Integer]: TSqDgMsg read GetMessage write SetMessage;
  end;
  
  TSequenceDiagram = class (TObject)
  private
    FDeltaStartEnd: Integer;
    FDeltaX: Integer;
    FDeltaY: Integer;
    FDiagram: IMMDiagram;
    FDiagramExplorer: IMMDiagramExplorer;
    FDiagramManager: IMMDiagramManager;
    FFirstMsgYPos: Integer;
    FMessages: TSqMsgList;
    FName: string;
    FParent: IMMDiagram;
    FRoles: TSqRoleList;
    FSequence: TStringList;
    FToolServices: IMMToolServices;
    procedure GetClassMemberNames(aLine : String; var aClassName,aMemberName : 
            String);
    function GetRole(aRoleName : String; var aRole : TSqDgRole): Boolean;
    procedure PositionActivations;
    procedure PositionMessages;
    procedure PositionRoles;
  public
    constructor Create(DiagramName : String; DiagramParent : IMMDiagram);
    destructor Destroy; override;
    procedure AssignSequence(aSequence : TStringList);
    procedure Commit;
    function CreateMessages: TMsgsInfo;
    function CreateRoles: Integer;
    procedure DoPositioning(dX,dY,dStartEnd : Integer);
    procedure SetPage(PageFormat : TPageFormat; PageOrientation : 
            TPageOrientation);
    property Diagram: IMMDiagram read FDiagram;
  end;
  
  TDiagramGenerator = class (TObject)
  private
    FDeltaStartEnd: Integer;
    FDeltaX: Integer;
    FDeltaY: Integer;
    FDiagnose: TDiagnose;
    FExMsg: string;
    FMsgsInfo: TMsgsInfo;
    FNumberOfRoles: Integer;
    FPageFormat: TPageFormat;
    FPageOrientation: TPageOrientation;
    FSequenceDiagram: TSequenceDiagram;
    function CalcPageFormat: TPageFormat;
    function CalcPageOrientation: TPageOrientation;
    procedure CalculatePositioning;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AssignDiag(aDiag : TDiagnose);
    procedure GenerateDefaultFolderSet;
    function GenSqDg(Parent : IMMDiagram; DiagramName : String;  Sequence: 
            TStringList): Integer;
    procedure InvertSequence(var aSequence : TStringList);
    function LocateClass(aClassName : String): IMMClassBase;
    procedure SelectAsCurrentDiagram;
  end;
  

procedure Register;

implementation

procedure Register;
begin
end;

{
********************************** TSqDgRole ***********************************
}
constructor TSqDgRole.Create(Parent : TSequenceDiagram);
begin
  FParent:=Parent;
  FSymbol:=FParent.FDiagramExplorer.CreateSymbol('rolesymbol', nil);
  FClassBase:=NIL;
end;

destructor TSqDgRole.Destroy;
begin
end;

procedure TSqDgRole.AddExternalDocHyperlink(aHyperlink : String);
begin
  FSymbol.AddExternalDocHyperlink(aHyperlink);
end;

function TSqDgRole.GetAutoSizeX: Boolean;
begin
  Result:=FSymbol.AutoSizeX;
end;

function TSqDgRole.GetAutoSizeY: Boolean;
begin
  Result:=FSymbol.AutoSizeY;
end;

function TSqDgRole.GetHeight: Integer;
begin
  Result:=FSymbol.Height;
end;

function TSqDgRole.GetInstanceName: string;
begin
  Result:=FSymbol.PropValue['InstanceName'];
end;

function TSqDgRole.GetName: string;
begin
  Result:=FSymbol.Name;
end;

function TSqDgRole.GetRoleKind: string;
begin
  Result:=FSymbol.PropValue['RoleKind'];
end;

function TSqDgRole.GetWidth: Integer;
begin
  Result:=FSymbol.Width;
end;

procedure TSqDgRole.HotLinkToEntity(aEntity : TLinkableEntity; aId : Integer);
begin
  FSymbol.HotlinkToEntity(aEntity,aId);
end;

procedure TSqDgRole.Move(xPos,yPos : integer);
begin
  FSymbol.Move(xPos,yPos);
  FPositionX:=xPos;
  FPositionY:=yPos;
end;

procedure TSqDgRole.SetAutoSizeX(Value: Boolean);
begin
  FSymbol.AutoSizeX:=Value;
end;

procedure TSqDgRole.SetAutoSizeY(Value: Boolean);
begin
  FSymbol.AutoSizeY:=Value;
end;

procedure TSqDgRole.SetHeight(Value: Integer);
begin
  FSymbol.Height:=Value;
end;

procedure TSqDgRole.SetInstanceName(const Value: string);
begin
  FSymbol.PropValue['InstanceName']:=Value;
end;

procedure TSqDgRole.SetName(const Value: string);
var
  i: Integer;
  aClass: IMMClassBase;
begin
  FSymbol.Name:=Value;
  // Im Model nach Klasse mit diesem Namen suchen
  for i:=0 to MMToolServices.CodeModel.ClassCount-1 do
  begin
    aClass:=MMToolServices.CodeModel.Classes[i];
    if aClass.Name=Value then
    begin
      HotlinkToEntity(nttClass,aClass.Id);
      FClassBase:=aClass;
      RoleKind:='rkInstance';
      exit;
    end;
  end;
  RoleKind:='rkActor';
end;

procedure TSqDgRole.SetRoleKind(const Value: string);
begin
  if ((Value='rkInstance') or
      (Value='rkActor') or
      (Value='rkSysBounds')
     ) then
  begin
    FSymbol.PropValue['RoleKind'] :=Value;
  end;
end;

procedure TSqDgRole.SetWidth(Value: Integer);
begin
  FSymbol.Width:=Value;
end;

{
*********************************** TSqDgMsg ***********************************
}
constructor TSqDgMsg.Create(Parent : TSequenceDiagram; Source,Target : 
        TSqDgRole);
begin
  FParent:=Parent;
  FSourceRole:=Source;
  FTargetRole:=Target;
end;

destructor TSqDgMsg.Destroy;
begin
end;

procedure TSqDgMsg.Commit;
var
  aMemberName: string;
  aPos: Integer;
  i: Integer;
  aMember: IMMMember;
begin
  FAssociation:=FParent.FDiagramExplorer.CreateAssociation('sequencemessage', FSourceRole.Symbol, FTargetRole.Symbol);
  FAssociation.PropValue['MessageKind']:=FMessageKind;
  FAssociation.PropValue['Condition'] :=FCondition;
  FAssociation.PropValue['StartTime'] :=FStartTime;
  FAssociation.PropValue['EndTime']   :=FEndTime;
  FAssociation.PropValue['Parameters']:=FParameters;
  
  //========= Name und Link
  FAssociation.Name:=FName;
  aMemberName:=FName;
  if aMemberName<>'' then
  begin
    aPos:=Pos('(',aMemberName);
    if aPos>0 then
      aMemberName:=Copy(aMemberName,1,aPos-1);
  end;
  //Message ist im allgemeinen vom Target die Methode...
  //also suchen wir sie und generieren ein entsprechnedes hotlink
  if assigned(FTargetRole.ClassBase) then
  begin
    for i:=0 to FTargetRole.ClassBase.MemberCount-1 do
    begin
      aMember := FTargetRole.ClassBase.Members[i];
      if aMember.Name=aMemberName then
      begin
        FAssociation.HotlinkToEntity(nttMember, aMember.ID);
      end;
    end;
  end;
  //=======================
  
  FAssociation.Move(0,StrToInt(FStartTime));
end;

function TSqDgMsg.GetEndTime: Integer;
begin
  Result:=StrToInt(FEndTime);
end;

function TSqDgMsg.GetMessageKind: string;
begin
  Result:=FMessageKind;
end;

function TSqDgMsg.GetName: string;
begin
  Result:=FName;
end;

function TSqDgMsg.GetParameters: string;
begin
  Result:=FParameters;
end;

function TSqDgMsg.GetStartTime: Integer;
begin
  Result:=StrToInt(FStartTime);
end;

procedure TSqDgMsg.SetCondition(const Value: string);
begin
  FCondition:=Value;
end;

procedure TSqDgMsg.SetEndTime(Value: Integer);
begin
  FEndTime:=IntToStr(Value);
end;

procedure TSqDgMsg.SetMessageKind(const Value: string);
begin
  FMessageKind:=Value;
end;

procedure TSqDgMsg.SetName(const Value: string);
begin
  FName:=Value;
end;

procedure TSqDgMsg.SetParameters(const Value: string);
begin
  FParameters:=Value;
end;

procedure TSqDgMsg.SetStartTime(const Value: Integer);
begin
  FStartTime:=IntToStr(Value);
end;

{
********************************* TSqRoleList **********************************
}
constructor TSqRoleList.Create;
begin
  inherited Create;
end;

function TSqRoleList.Add(Item: TSqDgRole): Integer;
begin
  Result := inherited Add(Item);
end;

function TSqRoleList.Extract(Item: TSqDgRole): TSqDgRole;
begin
  Result := TSqDgRole(inherited Extract(Item));
end;

function TSqRoleList.First: TSqDgRole;
begin
  Result := TSqDgRole(inherited First);
end;

function TSqRoleList.GetRole(Index: Integer): TSqDgRole;
begin
  Result:=TSqDgRole(Get(Index));
end;

function TSqRoleList.IndexOf(Item: TSqDgRole): Integer;
begin
  Result := inherited IndexOf(Item);
end;

procedure TSqRoleList.Insert(Index: Integer; Item: TSqDgRole);
begin
  inherited Insert(Index, Item);
end;

function TSqRoleList.Last: TSqDgRole;
begin
  Result := TSqDgRole(inherited Last);
end;

function TSqRoleList.Remove(Item: TSqDgRole): Integer;
begin
  Result := inherited Remove(Item);
end;

procedure TSqRoleList.SetRole(Index: Integer; Value: TSqDgRole);
begin
  Put(Index,Value);
end;

procedure TSqRoleList.Sort;
begin
  raise Exception.Create('Sort not done jet');
end;

{
********************************** TSqMsgList **********************************
}
constructor TSqMsgList.Create;
begin
  inherited Create;
end;

function TSqMsgList.Add(Item: TSqDgMsg): Integer;
begin
  Result := inherited Add(Item);
end;

function TSqMsgList.Extract(Item: TSqDgMsg): TSqDgMsg;
begin
  Result := TSqDgMsg(inherited Extract(Item));
end;

function TSqMsgList.First: TSqDgMsg;
begin
  Result := TSqDgMsg(inherited First);
end;

function TSqMsgList.GetMessage(Index: Integer): TSqDgMsg;
begin
  Result:=TSqDgMsg(Get(Index));
end;

function TSqMsgList.IndexOf(Item: TSqDgMsg): Integer;
begin
  Result := inherited IndexOf(Item);
end;

procedure TSqMsgList.Insert(Index: Integer; Item: TSqDgMsg);
begin
  inherited Insert(Index, Item);
end;

function TSqMsgList.Last: TSqDgMsg;
begin
  Result := TSqDgMsg(inherited Last);
end;

function TSqMsgList.Remove(Item: TSqDgMsg): Integer;
begin
  Result := inherited Remove(Item);
end;

procedure TSqMsgList.SetMessage(Index: Integer; Value: TSqDgMsg);
begin
  Put(Index,Value);
end;

procedure TSqMsgList.Sort;
begin
  raise Exception.Create('Sort not done jet');
end;

{
******************************* TSequenceDiagram *******************************
}
constructor TSequenceDiagram.Create(DiagramName : String; DiagramParent : 
        IMMDiagram);
var
  ParentId: Integer;
begin
  FParent:=DiagramParent;
  if assigned (FParent) then
    ParentId:=FParent.Id
  else
    ParentId:=0;
  
  FFirstMsgYPos:=40;
  FName:=DiagramName;
  FToolServices:=MMToolServices;
  FDiagramManager:=FToolServices.DiagramManager;
  FDiagram:=FDiagramManager.AddDiagram(diSequenceDiagram, DiagramName,ParentId);
  FDiagramExplorer := FDiagram.Explore as IMMDiagramExplorer;
  
  FRoles:=TSqRoleList.Create;
  FMessages:=TSqMsgList.Create;
  FSequence:=TStringList.Create;
end;

destructor TSequenceDiagram.Destroy;
var
  i: Integer;
begin
  for i:=0 to FRoles.Count-1 do
    FRoles.Items[i].Free;
  FRoles.Free;
  
  for i:=0 to FMessages.Count-1 do
    FMessages.Items[i].Free;
  FMessages.Free;
  
  FSequence.Free;
end;

procedure TSequenceDiagram.AssignSequence(aSequence : TStringList);
var
  i: Integer;
begin
  FSequence.Clear;
  
  for i:=0 to aSequence.Count-1 do
    FSequence.Add(aSequence.Strings[i]);
end;

procedure TSequenceDiagram.Commit;
var
  i: Integer;
begin
  for i:=0 to FMessages.Count-1 do
    FMessages.Items[i].Commit;
  
  PositionActivations;
  FDiagramExplorer.Commit;
end;

function TSequenceDiagram.CreateMessages: TMsgsInfo;
var
  RoleS: TSqDgRole;
  RoleT: TSqDgRole;
  DgMsg: TSqDgMsg;
  aRoleName: string;
  aMsgName: string;
  i: Integer;
begin
  // First Message
  //===========
  Result.NumberOfMessages:=0;
  Result.NumberOfSelfCalls:=0;
  
  GetClassMemberNames(FSequence.Strings[0],aRoleName,aMsgName);
  // validate results...
  if aRoleName='' then
    raise Exception.Create('empty Role name');
  
  // First Caller a ClassMessage???
  if aMsgName<>'' then
  begin
    GetRole('System',RoleS);
    // -> Destination Role
    GetRole(aRoleName,RoleT);
    DgMsg:=TSqDgMsg.Create(Self,RoleS,RoleT);
    Inc(Result.NumberOfMessages);
    DgMsg.Name:=aMsgName;
    DgMsg.MessageKind:= 'mkCall';
    FMessages.Add(DgMsg);
    RoleS:=RoleT;
  end
  else
  begin
    // erster aufrufer ist keine Klassenmethode....
    GetRole(aRoleName,RoleS);
  end;
  
  //=== The Rest of the Messages
  for i:=1 to FSequence.Count-1 do
  begin
    GetClassMemberNames(FSequence.Strings[i],aRoleName,aMsgName);
  
    if aRoleName='' then
      raise Exception.Create('empty Role name');
  
    GetRole(aRoleName,RoleT);
    DgMsg:=TSqDgMsg.Create(Self,RoleS,RoleT);
    Inc(Result.NumberOfMessages);
    if RoleS.Name=RoleT.Name then
      Inc(Result.NumberOfSelfCalls);
    DgMsg.Name:=aMsgName;
    DgMsg.MessageKind:= 'mkCall';
    FMessages.Add(DgMsg);
    RoleS:=RoleT;
  end;
end;

function TSequenceDiagram.CreateRoles: Integer;
var
  aRoleName: string;
  aMsgName: string;
  RoleS: TSqDgRole;
  RoleT: TSqDgRole;
  i: Integer;
begin
  // First Role
  //===========
  Result:=0;
  GetClassMemberNames(FSequence.Strings[0],aRoleName,aMsgName);
  
  // validate results...
  if aRoleName='' then
    raise Exception.Create('empty Role name');
  
  // First Caller a ClassMessage???
  if aMsgName<>'' then
  begin
    if GetRole('System',RoleS) then Inc(Result);
    RoleS.RoleKind:='rkSysBounds';
    // -> Destination Role
    if GetRole(aRoleName,RoleT) then Inc(Result);
  end
  else
  begin
    // erster aufrufer ist keine Klassenmethode....
    if GetRole(aRoleName,RoleS) then Inc(Result);
  end;
  
  //=== The Rest of the roles
  for i:=1 to FSequence.Count-1 do
  begin
    GetClassMemberNames(FSequence.Strings[i],aRoleName,aMsgName);
    if GetRole(aRoleName,RoleS) then Inc(Result);
  end;
end;

procedure TSequenceDiagram.DoPositioning(dX,dY,dStartEnd : Integer);
begin
  FDeltaX:=dx;
  FDeltaY:=dy;
  FDeltaStartEnd:=dStartEnd;
  PositionRoles;
  PositionMessages;
end;

procedure TSequenceDiagram.GetClassMemberNames(aLine : String; var aClassName,
        aMemberName : String);
var
  aPos: Integer;
begin
  aClassName:='';
  aMemberName:='';
  // extract Classname an Member
  aPos:=Pos('.',aLine);
  if aPos>0 then
  begin
    aClassName:=Copy(aLine,1,aPos-1);
    Delete(aLine,1,Length(aClassName)+1);
    aMemberName:=aLine;
  end
  else
  begin
    aClassName:=aLine;
    aMemberName:='';
  end;
end;

function TSequenceDiagram.GetRole(aRoleName : String; var aRole : TSqDgRole): 
        Boolean;
var
  i: Integer;
begin
  Result:=False;
  for i:=0 to FRoles.Count-1 do
  begin
    if FRoles.Items[i].Name=aRoleName then
    begin
      aRole:=FRoles.Items[i];
      exit;
    end;
  end;
  aRole:=TSqDgRole.Create(Self);
  aRole.Name:=aRoleName;
  FRoles.Add(aRole);
  Result:=True;
end;

procedure TSequenceDiagram.PositionActivations;
var
  i: Integer;
  MsgSource: TSqDgMsg;
  MsgTarget: TSqDgMsg;
  Role: IMMRoleSymbol;
  Activations: string;
begin
  //First Clear all activations...
  for i:=0 to FRoles.Count-1 do
    (FRoles.Items[i].Symbol as IMMRoleSymbol).Activations:='';
  
  MsgSource:=FMessages.Items[0];
  Role:=(MsgSource.SourceRole.Symbol as IMMRoleSymbol);
  Role.Activations:=Role.Activations+'('+IntToStr(MsgSource.StartTime-5)+','+IntToStr(MsgSource.StartTime+5)+')';
  
  for i:=1 to FMessages.Count-1 do
  begin
    MsgTarget:=FMessages.Items[i];
    Role:=(MsgTarget.SourceRole.Symbol as IMMRoleSymbol);
    Activations:=Role.Activations;
    Activations:=Activations+'('+IntToStr(MsgSource.EndTime-5)+','+IntToStr(MsgTarget.StartTime+5)+')';
    Role.Activations:=Activations;
    MsgSource:=MsgTarget;
  end;
  
  Role:=(MsgSource.TargetRole.Symbol as IMMRoleSymbol);
  Activations:=Role.Activations;
  Activations:=Activations+'('+IntToStr(MsgSource.EndTime-5)+','+IntToStr(MsgSource.EndTime+5)+')';
  Role.Activations:=Activations;
end;

procedure TSequenceDiagram.PositionMessages;
var
  i: Integer;
  yPos: Integer;
begin
  yPos:=FFirstMsgYPos;
  
  for i:=0 to FMessages.Count-1 do
  begin
    FMessages.Items[i].StartTime:=yPos;
    if FMessages.Items[i].SourceRole.Name=FMessages.Items[i].TargetRole.Name then
    begin
      yPos:=yPos+FDeltaY;
    end;
    FMessages.Items[i].EndTime:=yPos;
    yPos:=yPos+FDeltaY;
  end;
end;

procedure TSequenceDiagram.PositionRoles;
var
  i: Integer;
  xPos: Integer;
begin
  xPos:=0;
  for i:=0 to FRoles.Count-1 do
  begin
    FRoles.Items[i].Move(xPos,0);
    xPos:=xPos+FDeltaX;
  end;
end;

procedure TSequenceDiagram.SetPage(PageFormat : TPageFormat; PageOrientation : 
        TPageOrientation);
begin
  FDiagramExplorer.Page.SetPage(PageFormat,PageOrientation);
end;

{
****************************** TDiagramGenerator *******************************
}
constructor TDiagramGenerator.Create;
begin
  FDeltaX:=200;
  FDeltaY:=25;
  FDeltaStartEnd:=10;
end;

destructor TDiagramGenerator.Destroy;
begin
  // Nothing to do yet
end;

procedure TDiagramGenerator.AssignDiag(aDiag : TDiagnose);
begin
  FDiagnose:=aDiag;
end;

function TDiagramGenerator.CalcPageFormat: TPageFormat;
var
  y: Integer;
begin
  Result:=pfA4; // spter anhand der Nachrichten berechnen
  y:=FDeltaY*FMsgsInfo.NumberOfMessages+FDeltaY*FMsgsInfo.NumberOfSelfCalls;
  if y>1800 then
  begin
    Result:=pfA0;
    exit;
  end;
  if y>1400 then
  begin
    Result:=pfA1;
    exit;
  end;
  if y>1040 then
  begin
    Result:=pfA2;
    exit;
  end;
  if y>630 then
  begin
    Result:=pfA3;
    exit;
  end;
end;

function TDiagramGenerator.CalcPageOrientation: TPageOrientation;
begin
  Result:=poLandscape; // Spter anhand der Nachrichten berechnen
end;

procedure TDiagramGenerator.CalculatePositioning;
begin
  case FPageFormat of
    pfA4: FDeltaX:=1200 div FNumberOfRoles;
    pfA3: FDeltaX:=1600 div FNumberOfRoles;
    pfA2: FDeltaX:=2400 div FNumberOfRoles;
    pfA1: FDeltaX:=3000 div FNumberOfRoles;
    pfA0: FDeltaX:=4000 div FNumberOfRoles;
    else
      FDeltaX:=1200 div FNumberOfRoles;
  end;
end;

procedure TDiagramGenerator.GenerateDefaultFolderSet;
var
  ParentId: Integer;
  FolderList: TStringList;
  i: Integer;
  FolderListPath: string;
  CurDiagram: IMMDiagram;
begin
  CurDiagram:=MMToolServices.Navigator.CurDiagram;
  if assigned(CurDiagram) then
    ParentId:=CurDiagram.Id
  else
    ParentId:=0;
  
  FolderList:=TStringList.Create;
  try
    FolderListPath:=ExtractFilePath(Application.ExeName)+'DefaultFolder.txt';
    if FileExists(FolderListPath) then
    begin
      FolderList.LoadFromFile(FolderListPath);
    end
    else
    begin
      FolderList.Add('UseCases');
      FolderList.Add('Activities');
      FolderList.Add('Sequences');
      FolderList.Add('Classes');
      FolderList.Add('MindMaps');
      FolderList.SaveToFile(FolderListPath);
    end;
  
    if ParentId=0 then
    begin
      CurDiagram:=MMToolServices.DiagramManager.AddDiagram(diDiagramFolder,'NewProject', ParentId);
      ParentId:=CurDiagram.Id;
    end;
  
    for i:=0 to FolderList.Count-1 do
    begin
      MMToolServices.DiagramManager.AddDiagram(diDiagramFolder,FolderList.Strings[i], ParentId);
    end;
  
  finally
    FolderList.Free;
  end;
end;

function TDiagramGenerator.GenSqDg(Parent : IMMDiagram; DiagramName : String;  
        Sequence: TStringList): Integer;
begin
  //=== TRY
  Result:=-1;
  TRY
  
  FSequenceDiagram:=TSequenceDiagram.Create(DiagramName,Parent);
  try
    FSequenceDiagram.AssignSequence(Sequence);
    FNumberOfRoles:=FSequenceDiagram.CreateRoles;
    FMsgsInfo:=FSequenceDiagram.CreateMessages;
    FPageFormat:=CalcPageFormat;
    FPageOrientation:=CalcPageOrientation;
    FSequenceDiagram.SetPage(FPageFormat,FPageOrientation);
    CalculatePositioning;
    FSequenceDiagram.DoPositioning(FDeltaX,FDeltaY,FDeltaStartEnd);
    FSequenceDiagram.Commit;
    Result:=FSequenceDiagram.Diagram.Id;
  finally
    FSequenceDiagram.Free;
    FSequenceDiagram:=NIL;
  end;
  
  //===
  EXCEPT
    on E: Exception do
    begin
      FExMsg:='Exception in TDiagramGenerator.GenSqDg:'+E.Message;
      if FDiagnose<>nil then
        FDiagnose.Report('E',FExMsg)
      else
        ShowMessage(FExMsg);
    end;
  END;
end;

procedure TDiagramGenerator.InvertSequence(var aSequence : TStringList);
var
  aTmpList: TStringList;
  i: Integer;
begin
  aTmpList:=TStringList.Create;
  try
    for i:=aSequence.Count-1 downto 0 do
      aTmpList.Add(aSequence.Strings[i]);
    aSequence.Clear;
    for i:=0 to aTmpList.Count-1 do
      aSequence.Add(aTmpList.Strings[i]);
  finally
    aTmpList.Clear;
    aTmpList.Free;
  end;
end;

function TDiagramGenerator.LocateClass(aClassName : String): IMMClassBase;
var
  i: Integer;
  aClass: IMMClassBase;
begin
  result:=NIL;
  //=== check if a class with this name exists
  for i:=0 to MMToolServices.CodeModel.ClassCount-1 do
  begin
    aClass:=MMToolServices.CodeModel.Classes[i];
    if aClass.Name=aClassName then
    begin
      result:=aClass;
      exit;
    end;
  end;
end;

procedure TDiagramGenerator.SelectAsCurrentDiagram;
begin
end;


end.
