{------------------------------------------------------------------------------

Copyright (C) 2002  Rainer Queck
e-mail: Rainer.Queck@Qutronic.de
http://www.qutronic.de


This code is free; you can redistribute it and/or
modify it under the terms of the MPL MOZILLA PUBLIC LICENSE. 

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
MPL MOZILLA PUBLIC LICENSE for more details at http://www.mozilla.org/MPL

//----------------------------------------------------------------------------}
// Unit Created at    : somewhen in 2002
// Unit last modified : Date: 21.12.2002  Time :10:13:30
//------------------------------------------------------------------------------
unit DiagramGrouping;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  Forms, Dialogs,stdctrls,TypeStream,MMToolsApi,MMDiagramApi,MMEngineDefs,
  PrjNotify,RQClasses;

type
  TDGMember = class;
  TDiagramGroup = class;
  TDiagramGroupManager = class;
  TDGMemberList = class;
  TDiagramGroupList = class;
  TDGMember = class (TObject)
  private
    FDiagnose: TDiagnose;
    FDiagramId: Integer;
    FExMsg: string;
    FName: string;
    FPageNo: Integer;
    FValid: Boolean;
    function GetName: string;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AssignDiag(aDiag : TDiagnose);
    procedure LoadFromStream(aStream : TTypeMemoryStream);
    procedure Print;
    procedure SaveToStream(aStream : TTypeMemoryStream);
    property DiagramId: Integer read FDiagramId write FDiagramId;
    property Name: string read GetName;
    property PageNo: Integer read FPageNo write FPageNo;
    property Valid: Boolean read FValid write FValid;
  end;
  
  TDiagramGroup = class (TObject)
  private
    FMembers: TDGMemberList;
    FName: string;
  public
    constructor Create;
    destructor Destroy; override;
    function AddCurrentDiagram: Boolean;
    procedure LoadFromStream(aStream : TTypeMemoryStream);
    procedure SaveToStream(aStream : TTypeMemoryStream);
    property Members: TDGMemberList read FMembers write FMembers;
    property Name: string read FName write FName;
  end;
  
  TDiagramGroupManager = class (TObject)
  private
    FCurrentGroup: TDiagramGroup;
    FDiagramGroups: TDiagramGroupList;
    FGUIGroupSelect: TComboBox;
    FGUIMemberList: TListBox;
    FGUI_Initialized: Boolean;
    FProjectName: string;
    FProjectNotifierObserver: TProjectNotifierObserver;
    procedure AfterOpen(Sender : TObject);
    procedure BeforeOpen(Sender : TObject);
    function GroupAlreadyExist(aGroupName : String): Boolean;
    procedure GroupSelectChange(Sender: TObject);
    procedure GroupSelectDropDown(Sender: TObject);
    procedure MemberListClick(Sender : TObject);
    procedure MemberListKeyDown(Sender: TObject; var Key: Word; Shift: 
            TShiftState);
  public
    constructor Create;
    destructor Destroy; override;
    procedure AddCurrentDiagram;
    procedure AddGroup(aGroupName : String);
    procedure DeleteCurrentGroup;
    procedure DeleteDiagramFromGroup;
    procedure GetInvalidDGMembers(var aMemberList : TStringList);
    procedure InitializeGUI;
    procedure LoadFromFile(aPathFileName : String);
    procedure LoadFromStream(aStream : TTypeMemoryStream);
    procedure Save;
    procedure SaveToFile(aPathFileName : String);
    procedure SaveToStream(aStream : TTypeMemoryStream);
    property GUIGroupSelect: TComboBox read FGUIGroupSelect write 
            FGUIGroupSelect;
    property GUIMemberList: TListBox read FGUIMemberList write FGUIMemberList;
    property GUI_Initialized: Boolean read FGUI_Initialized;
    property ProjectName: string read FProjectName write FProjectName;
  end;
  
  TDGMemberList = class (TList)
  private
    function GetDGMember(Index: Integer): TDGMember;
    procedure SetDGMember(Index: Integer; Value: TDGMember);
  public
    constructor Create;
    function Add(Item: TDGMember): Integer;
    function Extract(Item: TDGMember): TDGMember;
    function First: TDGMember;
    function IndexOf(Item: TDGMember): Integer;
    procedure Insert(Index: Integer; Item: TDGMember);
    function Last: TDGMember;
    function Remove(Item: TDGMember): Integer;
    procedure Sort;
    property Items[Index: Integer]: TDGMember read GetDGMember write 
            SetDGMember;
  end;
  
  TDiagramGroupList = class (TList)
  private
    function GetDiagramGroup(Index: Integer): TDiagramGroup;
    procedure SetDiagramGroup(Index: Integer; Value: TDiagramGroup);
  public
    constructor Create;
    function Add(Item: TDiagramGroup): Integer;
    function Extract(Item: TDiagramGroup): TDiagramGroup;
    function First: TDiagramGroup;
    function IndexOf(Item: TDiagramGroup): Integer;
    procedure Insert(Index: Integer; Item: TDiagramGroup);
    function Last: TDiagramGroup;
    function Remove(Item: TDiagramGroup): Integer;
    procedure Sort;
    property Items[Index: Integer]: TDiagramGroup read GetDiagramGroup write 
            SetDiagramGroup;
  end;
  


implementation
uses QmmeMain;


{
********************************** TDGMember ***********************************
}
constructor TDGMember.Create;
begin
  inherited Create;
  
  FDiagramId := -1;
  FPageNo := -1;
  FValid := True;
end;

destructor TDGMember.Destroy;
begin
  
  inherited Destroy;
end;

procedure TDGMember.AssignDiag(aDiag : TDiagnose);
begin
  FDiagnose:=aDiag;
end;

function TDGMember.GetName: string;
var
  aIndex: Integer;
begin
  Result:='Invalid Id:'+IntToStr(FDiagramId);
  //LookupDiagram
  if MMToolServices.DiagramManager.FindDiagram(FDiagramId,aIndex) then
  begin
    Result:=MMToolServices.DiagramManager.Diagrams[aIndex].Name;
  end;
  FName:=Result;
end;

procedure TDGMember.LoadFromStream(aStream : TTypeMemoryStream);
var
  aName: string;
  aIndex: Integer;
  i: Integer;
begin
  FDiagramId:=aStream.ReadInteger;
  FName     :=aStream.ReadString;
  FPageNo   :=aStream.ReadInteger;
  
  if MMToolServices.DiagramManager.FindDiagram(FDiagramId,aIndex) then
  begin
    aName:=MMToolServices.DiagramManager.Diagrams[aIndex].Name;
    if aName<>FName then
    begin
      // Name <-> DiagramId Zuordnung reparieren
      for i:=0 to MMToolServices.DiagramManager.DiagramCount-1 do
      begin
        aName:=MMToolServices.DiagramManager.Diagrams[aIndex].Name;
        if aName<>FName then
        begin
          FDiagramId:=MMToolServices.DiagramManager.Diagrams[aIndex].Id;
          exit; // alles wieder OK
        end;
      end;
      FValid:=False;
    end;
  end;
end;

procedure TDGMember.Print;
begin
  //=== TRY
  TRY
  
  // use Pageno to print! >>> Seite X von Y oder so
  raise Exception.Create('ToDo');
  
  //===
  EXCEPT
    on E: Exception do
    begin
      FExMsg:='Exception in TDGMember.Print:'+E.Message;
      if FDiagnose<>nil then
        FDiagnose.Report('E',FExMsg)
      else
        ShowMessage(FExMsg);
    end;
  END;
end;

procedure TDGMember.SaveToStream(aStream : TTypeMemoryStream);
begin
  aStream.WriteInteger(FDiagramId);
  aStream.WriteString(FName);
  aStream.WriteInteger(FPageNo);
end;

{
******************************** TDiagramGroup *********************************
}
constructor TDiagramGroup.Create;
begin
  inherited Create;
  
  FMembers:=TDGMemberList.Create;
  FName:='UnNamed';
end;

destructor TDiagramGroup.Destroy;
var
  i: Integer;
begin
  for i:=0 to FMembers.Count-1 do
  begin
    FMembers.Items[i].Free;
    FMembers.Items[i]:=NIL;
  end;
  FMembers.Free;
  
  inherited Destroy;
end;

function TDiagramGroup.AddCurrentDiagram: Boolean;
var
  i: Integer;
  aDGMember: TDGMember;
begin
  Result:=False;
  for i:=0 to FMembers.Count-1 do
  begin
    if FMembers.Items[i].DiagramId=MMToolServices.Navigator.CurDiagram.Id then
    begin
       // Diagram is already a member
       exit;
    end;
  end;
  aDGMember:=TDGMember.Create;
  aDGMember.DiagramId:=MMToolServices.Navigator.CurDiagram.Id;
  FMembers.Add(aDGMember);
  Result:=True;
end;

procedure TDiagramGroup.LoadFromStream(aStream : TTypeMemoryStream);
var
  MemberCount: Integer;
  i: Integer;
  aMember: TDGMember;
begin
  FName:=aStream.ReadString;
  MemberCount:=aStream.ReadInteger;
  for i:=0 to MemberCount-1 do
  begin
    aMember:=TDGMember.Create;
    aMember.LoadFromStream(aStream);
    FMembers.Add(aMember);
  end;
end;

procedure TDiagramGroup.SaveToStream(aStream : TTypeMemoryStream);
var
  i: Integer;
begin
  aStream.WriteString(FName);
  aStream.WriteInteger(FMembers.Count);
  
  //=== Memers Streamen sich selbst...
  for i:=0 to FMembers.Count-1 do
    FMembers.Items[i].SaveToStream(aStream);
end;

{
***************************** TDiagramGroupManager *****************************
}
constructor TDiagramGroupManager.Create;
begin
  inherited Create;
  
  FGUIGroupSelect:=NIL;
  FGUIMemberList:=NIL;
  FDiagramGroups:=TDiagramGroupList.Create;
  FCurrentGroup:=NIL;
  FProjectNotifierObserver:=NIL;
end;

destructor TDiagramGroupManager.Destroy;
var
  i: Integer;
begin
  inherited Destroy;
  
  if Assigned(FProjectNotifierObserver) then
  begin
    if assigned(ProjectNotifier) then
      ProjectNotifier.UnRegisterObserver(FProjectNotifierObserver);
    FProjectNotifierObserver.Free;
  end;
  if FProjectName<>'' then
    SaveToFile(FProjectName);
  for i:=0 to FDiagramGroups.Count-1 do
  begin
    FCurrentGroup:=FDiagramGroups.Items[i];
    FCurrentGroup.Free;
    FCurrentGroup:=NIL;
  end;
  FDiagramGroups.Free;
end;

procedure TDiagramGroupManager.AddCurrentDiagram;
begin
  if assigned(FCurrentGroup) then
  begin
    if FCurrentGroup.AddCurrentDiagram then
    begin
      FGUIMemberList.Items.Add(MMToolServices.Navigator.CurDiagram.Name);
    end;
  end
  else
  begin
    ShowMessage('No diagram group assigned yet');
  end;
end;

procedure TDiagramGroupManager.AddGroup(aGroupName : String);
begin
  if not GroupAlreadyExist(aGroupName) then
  begin
    FCurrentGroup:=TDiagramGroup.Create;
    FDiagramGroups.Add(FCurrentGroup);
    FCurrentGroup.Name:=aGroupName;
    FGUIGroupSelect.Text:=FCurrentGroup.Name;
    FGUIGroupSelect.Items.Add(FCurrentGroup.Name);
    FGUIMemberList.Clear;
  
    // and save the changes...
    SaveToFile(FProjectName);
  end;
end;

procedure TDiagramGroupManager.AfterOpen(Sender : TObject);
begin
  InitializeGUI;
end;

procedure TDiagramGroupManager.BeforeOpen(Sender : TObject);
var
  i: Integer;
begin
  Save;
  // unload all groups
  for i:=0 to FDiagramGroups.Count-1 do
  begin
    FCurrentGroup:=FDiagramGroups.Items[i];
    FCurrentGroup.Free;
    FCurrentGroup:=NIL;
    FDiagramGroups.Items[i]:=NIL;
  end;
  FDiagramGroups.Pack;
  FGUIGroupSelect.Items.Clear;
  //ProjektName bernehmen
  FProjectName:=MMToolServices.ProjectManager.ProjectName;
  FProjectName:=ChangeFileExt(FProjectName,'.grp');
end;

procedure TDiagramGroupManager.DeleteCurrentGroup;
var
  aIdx: Integer;
begin
  if MessageDlg('Please confirm deletion of current diagram group', mtConfirmation, [mbYes, mbNo], 0)=mrYes then
  begin
    aIdx:=FGUIGroupSelect.Items.IndexOf(FCurrentGroup.Name);
    if aIdx<0 then exit;
    FGUIGroupSelect.Items.Delete(aIdx);
    FDiagramGroups.Remove(FCurrentGroup);
    FDiagramGroups.Pack;
    FCurrentGroup.Free;
    if FDiagramGroups.Count>0 then
    begin
      FCurrentGroup:=FDiagramGroups.Items[0];
      FGUIGRoupSelect.Text:=FCurrentGroup.Name;
      GroupSelectChange(self);
    end;
  end;
end;

procedure TDiagramGroupManager.DeleteDiagramFromGroup;
var
  aDGName: string;
  i: Integer;
begin
  if MessageDlg('Please confirm deletion of selected diagram from current group', mtConfirmation, [mbYes, mbNo], 0)=mrYes then
  begin
    aDGName:=FGUIMemberList.Items[FGUIMemberList.ItemIndex];
    FGUIMemberList.Items.Delete(FGUIMemberList.ItemIndex);
    for i:=0 to FCurrentGroup.Members.Count-1 do
    begin
      if FCurrentGroup.Members.Items[i].Name=aDGName then
      begin
        FCurrentGroup.Members.Items[i].Free;
        FCurrentGroup.Members.Items[i]:=NIL;
      end;
    end;
    FCurrentGroup.Members.Pack;
  end;
end;

procedure TDiagramGroupManager.GetInvalidDGMembers(var aMemberList : 
        TStringList);
var
  i: Integer;
  k: Integer;
begin
  aMemberList.Clear;
  for i:=0 to FDiagramGroups.Count-1 do
    for k:=0 to FDiagramGroups.Items[i].Members.Count-1 do
    begin
      if FDiagramGroups.Items[i].Members.Items[k].Valid then continue;
      aMemberList.Add(FDiagramGroups.Items[i].Members.Items[k].Name);
    end;
end;

function TDiagramGroupManager.GroupAlreadyExist(aGroupName : String): Boolean;
var
  i: Integer;
begin
  Result:=False;
  for i:=0 to FDiagramGroups.Count-1 do
  begin
    if FDiagramGroups.Items[i].Name=aGroupName then
    begin
      Result:=True;
      exit;
    end;
  end;
end;

procedure TDiagramGroupManager.GroupSelectChange(Sender: TObject);
var
  i: Integer;
  aGroup: TDiagramGroup;
begin
  if not assigned(FDiagramGroups) then
  begin
    ShowMessage('No diagram groups assigned yet');
    exit;
  end;
  if not assigned(FCurrentGroup) then
  begin
    ShowMessage('No diagram groups assigned yet');
    exit;
  end;
  
  for i:=0 to FDiagramGroups.Count-1 do
  begin
    aGroup:=FDiagramGroups.Items[i];
    if aGroup.Name=FGUIGroupSelect.Text then
    begin
      FCurrentGroup:=aGroup;
      break;
    end;
  end;
  
  FGUIMemberList.Items.Clear;
  for i:=0 to FCurrentGroup.Members.Count-1 do
    FGUIMemberList.Items.Add(FCurrentGroup.Members.Items[i].Name);
end;

procedure TDiagramGroupManager.GroupSelectDropDown(Sender: TObject);
begin
end;

procedure TDiagramGroupManager.InitializeGUI;
var
  i: Integer;
begin
  //Create and register Observer
  if not assigned(FProjectNotifierObserver) then
  begin
    FProjectNotifierObserver:=TProjectNotifierObserver.Create(NIL);
    ProjectNotifier.RegisterObserver(FProjectNotifierObserver);
    FProjectNotifierObserver.OnBeforeOpen:=BeforeOpen;
    FProjectNotifierObserver.OnAfterOpen:=AfterOpen;
    FProjectNotifierObserver.Enabled:=True;
  end;
  
  FProjectName:=MMToolServices.ProjectManager.ProjectName;
  FProjectName:=ChangeFileExt(FProjectName,'.grp');
  
  with FGUIMemberList do
  begin
    OnKeyDown:=MemberListKeyDown;
    OnClick:=MemberListClick;
  end;
  with FGUIGroupSelect do
  begin
    OnChange:=GroupSelectChange;
    OnDropDown:=GroupSelectDropDown;
  end;
  
  FGUIMemberList.Clear;
  FGUIGroupSelect.Clear;
  for i:=0 to FDiagramGroups.Count-1 do
  begin
    FDiagramGroups.Items[i].Free;
    FDiagramGroups.Items[i]:=NIL;
  end;
  FDiagramGroups.Pack;
  
  if FileExists(FProjectName) then
    LoadFromFile(FProjectName);
  
  if FDiagramGroups.Count>0 then
  begin
    FCurrentGroup:=FDiagramGroups.Items[0];
    with FGUIMemberList do
    begin
      FGUIGroupSelect.Text:=FCurrentGroup.Name;
      for i:=0 to FCurrentGroup.Members.Count-1 do
      begin
        Items.Add(FCurrentGroup.Members.Items[i].Name);
      end;
    end;
  end;
  FGUI_Initialized:=True;
end;

procedure TDiagramGroupManager.LoadFromFile(aPathFileName : String);
var
  aStream: TTypeMemoryStream;
begin
  aStream:=TTypeMemoryStream.Create;
  try
    aStream.LoadFromFile(aPathFileName);
    aStream.Position:=0;
    LoadFromStream(aStream);
  finally
    aStream.Free;
  end;
end;

procedure TDiagramGroupManager.LoadFromStream(aStream : TTypeMemoryStream);
var
  i: Integer;
  aDiagramGroup: TDiagramGroup;
  GroupCount: Integer;
begin
  // unload all groups
  for i:=0 to FDiagramGroups.Count-1 do
  begin
    FCurrentGroup:=FDiagramGroups.Items[i];
    FCurrentGroup.Free;
    FCurrentGroup:=NIL;
    FDiagramGroups.Items[i]:=NIL;
  end;
  FDiagramGroups.Pack;
  FGUIGroupSelect.Items.Clear;
  
  GroupCount:=aStream.ReadInteger;
  for i:=0 to GroupCount-1 do
  begin
    aDiagramGroup:=TDiagramGroup.Create;
    aDiagramGroup.LoadFromStream(aStream);
    FDiagramGroups.Add(aDiagramGroup);
    FGUIGroupSelect.Items.Add(aDiagramGroup.Name);
  end;
end;

procedure TDiagramGroupManager.MemberListClick(Sender : TObject);
var
  aDGName: string;
  i: Integer;
  aDiagram: IMMDiagram;
  aIdx: Integer;
begin
  aDGName:=FGUIMemberList.Items[FGUIMemberList.ItemIndex];
  for i:=0 to FCurrentGroup.Members.Count-1 do
  begin
    if FCurrentGroup.Members.Items[i].Name=aDGName then
    begin
      if MMToolServices.DiagramManager.FindDiagram(FCurrentGroup.Members.Items[i].DiagramId,aIdx) then
      begin
        aDiagram:=MMToolServices.DiagramManager.Diagrams[aIdx];
        MMToolServices.Navigator.SetCurDiagram(aDiagram);
        MMToolServices.Navigator.ActivateView(vtDiagramEditor,True);
        exit;
      end;
    end;
  end;
end;

procedure TDiagramGroupManager.MemberListKeyDown(Sender: TObject; var Key: Word;
        Shift: TShiftState);
begin
  case Key of
    VK_INSERT : AddCurrentDiagram;
    VK_DELETE : DeleteDiagramFromGroup;
    //VK_UP     : FGUIMemberList.Items.Add('Up');
    //VK_DOWN   : FGUIMemberList.Items.Add('Down');
  end;
  Key:=0;
end;

procedure TDiagramGroupManager.Save;
begin
  if FileExists('.grp') then
    DeleteFile('.grp');
  
  SaveToFile(FProjectName);
end;

procedure TDiagramGroupManager.SaveToFile(aPathFileName : String);
var
  aStream: TTypeMemoryStream;
begin
  aStream:=TTypeMemoryStream.Create;
  try
    SaveToStream(aStream);
    aStream.Position:=0;
    aStream.SaveToFile(aPathFileName);
  finally
    aStream.Free;
  end;
end;

procedure TDiagramGroupManager.SaveToStream(aStream : TTypeMemoryStream);
var
  GroupCount: Integer;
  i: Integer;
begin
  //=== Anzahl der Gruppen festhalten
  GroupCount:=FDiagramGroups.Count;
  aStream.WriteInteger(GroupCount);
  
  //=== Die Einzelnen Gruppen streamen sich selbst....
  for i:=0 to FDiagramGroups.Count-1 do
    FDiagramGroups.Items[i].SaveToStream(aStream);
end;

{
******************************** TDGMemberList *********************************
}
constructor TDGMemberList.Create;
begin
  inherited Create;
end;

function TDGMemberList.Add(Item: TDGMember): Integer;
begin
  Result := inherited Add(Item);
end;

function TDGMemberList.Extract(Item: TDGMember): TDGMember;
begin
  Result := TDGMember(inherited Extract(Item));
end;

function TDGMemberList.First: TDGMember;
begin
  Result := TDGMember(inherited First);
end;

function TDGMemberList.GetDGMember(Index: Integer): TDGMember;
begin
  Result:=TDGMember(Get(Index));
end;

function TDGMemberList.IndexOf(Item: TDGMember): Integer;
begin
  Result := inherited IndexOf(Item);
end;

procedure TDGMemberList.Insert(Index: Integer; Item: TDGMember);
begin
  inherited Insert(Index, Item);
end;

function TDGMemberList.Last: TDGMember;
begin
  Result := TDGMember(inherited Last);
end;

function TDGMemberList.Remove(Item: TDGMember): Integer;
begin
  Result := inherited Remove(Item);
end;

procedure TDGMemberList.SetDGMember(Index: Integer; Value: TDGMember);
begin
  Put(Index,Value);
end;

procedure TDGMemberList.Sort;
begin
  raise Exception.Create('Sort not done jet');
end;

{
****************************** TDiagramGroupList *******************************
}
constructor TDiagramGroupList.Create;
begin
  inherited Create;
end;

function TDiagramGroupList.Add(Item: TDiagramGroup): Integer;
begin
  Result := inherited Add(Item);
end;

function TDiagramGroupList.Extract(Item: TDiagramGroup): TDiagramGroup;
begin
  Result := TDiagramGroup(inherited Extract(Item));
end;

function TDiagramGroupList.First: TDiagramGroup;
begin
  Result := TDiagramGroup(inherited First);
end;

function TDiagramGroupList.GetDiagramGroup(Index: Integer): TDiagramGroup;
begin
  Result:=TDiagramGroup(Get(Index));
end;

function TDiagramGroupList.IndexOf(Item: TDiagramGroup): Integer;
begin
  Result := inherited IndexOf(Item);
end;

procedure TDiagramGroupList.Insert(Index: Integer; Item: TDiagramGroup);
begin
  inherited Insert(Index, Item);
end;

function TDiagramGroupList.Last: TDiagramGroup;
begin
  Result := TDiagramGroup(inherited Last);
end;

function TDiagramGroupList.Remove(Item: TDiagramGroup): Integer;
begin
  Result := inherited Remove(Item);
end;

procedure TDiagramGroupList.SetDiagramGroup(Index: Integer; Value: 
        TDiagramGroup);
begin
  Put(Index,Value);
end;

procedure TDiagramGroupList.Sort;
begin
  raise Exception.Create('Sort not done jet');
end;


end.
