{------------------------------------------------------------------------------

Copyright (C) 2002  Rainer Queck
e-mail: Rainer.Queck@Qutronic.de
http://www.qutronic.de


This code is free; you can redistribute it and/or
modify it under the terms of the MPL MOZILLA PUBLIC LICENSE. 

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
MPL MOZILLA PUBLIC LICENSE for more details at http://www.mozilla.org/MPL

//---------------------------------------------------------------------------}
// Unit Created at    : Date: 28. November 1999
// Unit last modified : Date: 21.12.2002  Time :10:13:30
//------------------------------------------------------------------------------
unit MailboxRq;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  SyncObjs;

{$IFDEF VER100}
// Delphi3 -> windows.pas does not have this definitions
const
  SECURITY_DESCRIPTOR_REVISION = 1;
  SECURITY_NULL_SID_AUTHORITY    = 0;
  SECURITY_WORLD_SID_AUTHORITY   = 1;
  SECURITY_LOCAL_SID_AUTHORITY   = 2;
  SECURITY_CREATOR_SID_AUTHORITY = 3;
  SECURITY_NT_AUTHORITY = 5;
  SECURITY_WORLD_RID = 0;
  ACL_REVISION = 2;
  ACCESS_ALLOWED_ACE_SIZE = 12;
{$ENDIF}
type
  TSecurityEverybody = record
    sd: TSecurityDescriptor;
    sa: TSecurityAttributes;
    sia: TSIDIdentifierAuthority;
    acl: PACL;
    acllen: Integer;
    sid: pointer;
  end;
  TMailBoxErrors=(EStreamToLarge,ENoReceiverDefined,ENoReceiverExists,ESendError,
                  ERecSlotCreate);
  TMailBoxReceiverState=(srcvUnknown,srcvAvailable,srcvNotAvailable);

  TRqSlotThread = class;
  TRqMailBox = class;
  TMailslotRxEvent = procedure (Sender : TObject ; RxMessage : TMemoryStream) 
          of object;
  TMailSlotErrorEvent = procedure (Sender : TObject ; Error : TMailBoxErrors) 
          of object;
  TRqSlotThread = class (TThread)
  private
    FMaxBuffSize: Integer;
    FOnReceived: TMailslotRxEvent;
    FParent: TRqMailbox;
    FRcvBuff: TMemoryStream;
    FUsedByVcl: Boolean;
    procedure CallOnReceived;
  protected
    procedure Execute; override;
  public
    FSlotHandle: THandle;
    constructor Create(aParent : TRqMailbox; aOnReceived: TMailslotRxEvent; 
            UsedByVcl : Boolean);
    destructor Destroy; override;
    property MaxBuffSize: Integer read FMaxBuffSize write FMaxBuffSize;
  end;
  
  TRqMailBox = class (TObject)
  private
    FActive: Boolean;
    FExMsg: string;
    FMaxBuffSize: Integer;
    FOnError: TMailSlotErrorEvent;
    FOnMsgReceived: TMailslotRxEvent;
    FReceiver: string;
    FSecurityEverybody: TSecurityEverybody;
    FSlotHandle: THandle;
    FSlotName: string;
    FSlotThread: TRqSlotThread;
    FUsedByVCL: Boolean;
    function GetActive: Boolean;
    function GetMaxBuffSize: Integer;
    function GetReceiverAvailable: Boolean;
    function GetSlotName: string;
    procedure OnRxData(Sender : TObject ; RxMessage : TMemoryStream);
    procedure SetActive(Value: Boolean);
    procedure SetMaxBuffSize(Value: Integer);
    procedure SetSlotName(const Value: string);
  public
    constructor Create(UsedByVCL : Boolean);
    destructor Destroy; override;
    procedure Send(AMessage : TMemoryStream);
    property ReceiverAvailable: Boolean read GetReceiverAvailable;
  published
    property Active: Boolean read GetActive write SetActive;
    property BoxName: string read GetSlotName write SetSlotName;
    property MaxBuffSize: Integer read GetMaxBuffSize write SetMaxBuffSize;
    property OnError: TMailSlotErrorEvent read FOnError write FOnError;
    property OnReceive: TMailslotRxEvent read FOnMsgReceived write 
            FOnMsgReceived;
    property Receiver: string read FReceiver write FReceiver;
  end;
  

var
  RqMbCriticalSection : TCriticalSection;

implementation
{
********************************** TRqMailBox **********************************
}
constructor TRqMailBox.Create(UsedByVCL : Boolean);
begin
  inherited Create;
  
  with FSecurityEverybody do
  begin
    sa.nLength := sizeof (sa);
    InitializeSecurityDescriptor (@SD, SECURITY_DESCRIPTOR_REVISION);
    SetSecurityDescriptorDacl(@SD, TRUE, nil, TRUE);
    SetSecurityDescriptorGroup(@SD, nil, TRUE);
    SetSecurityDescriptorOwner(@SD, nil, TRUE);
    SA.lpSecurityDescriptor := @SD;
    SA.bInheritHandle := TRUE;
  end;
  FUsedByVCL:=UsedByVCL;
  FActive:=False;
  FReceiver:='';
  FSlotHandle:=INVALID_HANDLE_VALUE;
  MaxBuffSize:=1024;
end;

destructor TRqMailBox.Destroy;
begin
  if FActive then
    SetActive(False);
  
  with FSecurityEverybody do
  begin
    if Assigned(sid) then FreeSid(sid);
    if Assigned(acl) then Freemem(acl);
  end;
  inherited destroy;
end;

function TRqMailBox.GetActive: Boolean;
begin
  result:=FActive;
end;

function TRqMailBox.GetMaxBuffSize: Integer;
begin
  result:=FMaxBuffSize;
end;

function TRqMailBox.GetReceiverAvailable: Boolean;
var
  ReceiverHandle: THandle;
begin
  Result:=False;
  // lets see if we can open the send slot....
  ReceiverHandle:= CreateFile(PChar('\\.\mailslot\'+FReceiver),
                              GENERIC_WRITE,
                              FILE_SHARE_READ,
                              nil,
                              OPEN_ALWAYS,
                              FILE_ATTRIBUTE_NORMAL,
                              INVALID_HANDLE_VALUE);
  // if handle is valid...
  if not (ReceiverHandle=INVALID_HANDLE_VALUE) then
  begin
    Result:=True;
    CloseHandle(ReceiverHandle);
  end;
end;

function TRqMailBox.GetSlotName: string;
begin
  result:=FSlotName;
end;

procedure TRqMailBox.OnRxData(Sender : TObject ; RxMessage : TMemoryStream);
begin
  if assigned(FOnMsgReceived) then FOnMsgReceived(Sender,RxMessage);
end;

procedure TRqMailBox.Send(AMessage : TMemoryStream);
var
  ReceiverHandle: THandle;
  NumberOfBytesWritten: DWORD;
begin
  TRY
    RqMbCriticalSection.Enter;
  
    if FReceiver='' then
    begin
       if assigned(FOnError) then
         FOnError(self,ENoReceiverDefined)
       else
         raise Exception.Create('No receiver defined');
       exit;
    end;
    // lets open the send slot....
    ReceiverHandle:= CreateFile(PChar('\\.\mailslot\'+FReceiver),
                                GENERIC_WRITE,
                                FILE_SHARE_READ,
                                nil,
                                OPEN_ALWAYS,
                                FILE_ATTRIBUTE_NORMAL,
                                INVALID_HANDLE_VALUE);
    if (ReceiverHandle=INVALID_HANDLE_VALUE) then
    begin
      if assigned(FOnError) then
         FOnError(self,ENoReceiverExists)
      else
        raise Exception.Create('Receiver does not exist');
      exit;
    end;
  TRY
    // now send the stream ....
    AMessage.Position:=0;
    if not WriteFile(ReceiverHandle,AMessage.Memory^,AMessage.Size,NumberOfBytesWritten,nil) then
    begin
      if assigned(FOnError) then
         FOnError(self,ESendError)
      else
         raise Exception.Create('Send Error');
    end;
  FINALLY
    CloseHandle(ReceiverHandle);
  END;
  
  FINALLY
    RqMbCriticalSection.Leave;
  END;
end;

procedure TRqMailBox.SetActive(Value: Boolean);
var
  i: Integer;
begin
  //=== TRY
  TRY
  
  // if to mailslot is to close....
  if FActive and not Value then
  begin
    // Terminate SlotRcvThread
    FSlotThread.Terminate;
    if FSlotThread.Suspended then
      FSlotThread.Resume;
    i := WaitForSingleObject(FSlotThread.Handle, 10000);
    if i = WAIT_TIMEOUT then
    begin
      TerminateThread(FSlotThread.Handle, 0);
      ShowMessage('hat to kill SlotThread');
    end
    else
      FSlotThread.Free;
  
    CloseHandle(FSlotHandle);
    FSlotHandle:=INVALID_HANDLE_VALUE;
    FActive:=False;
  end
  // if mailslot is to open
  else if not FActive and Value then
  begin
    // Mail Slot einrichten...
    FSlotHandle := CreateMailSlot(PChar('\\.\mailslot\'+FSlotName),
                                  0,      // Any Size
                                  1000,   // Wait at least one second for a message
                                  @FSecurityEverybody.sa);
    if FSlotHandle=INVALID_HANDLE_VALUE then
    begin
      if assigned(FOnError) then
         FOnError(self,ERecSlotCreate)
      else
      raise Exception.Create('can not create Mailslot '+ FSlotName);
    end
    else
    begin
        // Create and Start RcvThread
       FSlotThread:=TRqSlotThread.Create(self,OnRxData,FUsedByVCL);
       FSlotThread.FSlotHandle:=FSlotHandle;
       FSlotThread.MaxBuffSize:=FMaxBuffSize;
       FSlotThread.Resume;
       FActive:=True;
    end;
  end;
  
  //===
  EXCEPT
    on E: Exception do
    begin
      FExMsg:=E.ClassName+': Exception in TRqMailBox.SetActive:'+E.Message;
      ShowMessage(FExMsg);
    end;
  END;
end;

procedure TRqMailBox.SetMaxBuffSize(Value: Integer);
begin
  // if not changed then don't care
  if Value=FMaxBuffSize then exit;
  
  FMaxBuffSize:=Value;
  if FActive then
  begin
    Active:=False;
    Active:=True;
  end;
end;

procedure TRqMailBox.SetSlotName(const Value: string);
begin
  if Value <> FSlotName then
  begin
    if FActive then
    begin
      SetActive(False);
      FSlotName:=Value;
      SetActive(True);
    end
    else FSlotName:=Value;
  end;
end;

{
******************************** TRqSlotThread *********************************
}
constructor TRqSlotThread.Create(aParent : TRqMailbox; aOnReceived: 
        TMailslotRxEvent; UsedByVcl : Boolean);
begin
  Inherited Create(True);
  FParent:=aParent;
  FUsedByVcl:=UsedByVcl;
  FOnReceived:=AOnReceived;
  FRcvBuff:=TMemoryStream.Create;
end;

destructor TRqSlotThread.Destroy;
begin
  FRcvBuff.Free;
  inherited destroy;
end;

procedure TRqSlotThread.CallOnReceived;
begin
  if (not Terminated) and Assigned(FOnReceived) then FOnReceived(FParent,FRcvBuff);
end;

procedure TRqSlotThread.Execute;
var
  i: DWORD;
  b: Boolean;
begin
  repeat
    FRcvBuff.Position:=0;
    FRcvBuff.SetSize(FMaxBuffsize);
    if FSlotHandle=INVALID_HANDLE_VALUE then
    begin
      sleep(100);
    end
    else
    begin
      b:=ReadFile(FSlotHandle, FRcvBuff.Memory^, FMaxBuffsize , i, nil);
      if b then
      begin
        FRcvBuff.Position:=0;
        FRcvBuff.SetSize(i);
        if FUsedByVcl then
          Synchronize(CallOnReceived)
        else
          CallOnReceived;
      end;
    end;
  until terminated;
end;


initialization
  RqMbCriticalSection:=TCriticalSection.Create;
finalization
  RqMbCriticalSection.Free;
end.

